(* self-application, functors, simple dispatch *)


use "../lib/smlprelude";


(* root objects ------------------------------------------------------- *)
 
open Closed;   (* closed datatype, % : closed *)

signature ROOT =
sig
    type 'a subself
    type 'a ext = 'a subself
    type self = closed subself
    val pack : unit -> ('a subself -> 'a subself) * ('a -> 'a ext)
    val unpack : 'a subself -> ('a -> 'a subself) * 'a
    val new : unit -> self
end;

structure Root :> ROOT =
struct

(* types *)
type 'a subself = 'a;
type self = closed subself;
type 'a ext = 'a subself;

(* packing *)
fun unpack ob = (I,ob);
fun pack () = (I,I);

(* new *)
fun new () = %;
end;

signature OBJECT =
sig
    type 'a ext
    type 'a subself
    type self = closed subself
    val unpack : 'a subself -> ('a -> 'a subself) * 'a
end;


(* eqobjs -------------------------------------------------------------	*)

signature EQUAL =
sig
    structure Super : OBJECT
    type 'a ext
    type 'a subself = 'a ext Super.subself
    type self = closed subself
    val pack : unit -> ('a subself -> 'a subself) * ('a -> 'a ext)
    val unpack : 'a subself -> ('a -> 'a subself) * 'a
    val repack : (('a subself -> 'a subself -> bool) ->
		  ('a subself -> 'a subself -> bool)) ->
		 'a subself -> 'a subself
    val eq : 'a subself -> 'a subself -> bool
end;

functor Equality (structure S : OBJECT) :> EQUAL where Super = S =
struct

structure Super = S;

(* types *)
datatype 'a state = S of
		'a subself -> 'a subself -> bool		(* eq	 *)
withtype 'a ext = 'a state * 'a
and 'a subself  = 'a ext Super.subself;
type self = closed subself;

(* messages *)
exception Eq;
fun eq p q = let val (_,(S e,_)) = Super.unpack p in e p q end;

(* packing *)
fun unpack p = explode (Super.unpack p);
fun repack f p = implode (fn S m => S (f m)) (Super.unpack p);
fun oeq p q = raise Eq;
fun pack () = (I, fn ext => (S oeq,ext));
end;

signature EQOBJECT =
    sig
	type 'a subself
	type self = closed subself
	val unpack : 'a subself -> ('a -> 'a subself) * 'a
	val repack_eq : (('a subself -> 'a subself -> bool) ->
		         ('a subself -> 'a subself -> bool)) ->
		        'a subself -> 'a subself
    end;


(* points -------------------------------------------------------------	*)

signature PLACED =
sig
    structure Super : EQOBJECT
    type 'a ext
    type 'a subself = 'a ext Super.subself
    type self = closed subself
    val pack : int -> ('a subself -> 'a subself) * ('a -> 'a ext)
    val unpack : 'a subself -> ('a -> 'a subself) * 'a
    val repack : (('a subself -> int) *
		  ('a subself -> 'a subself -> real) *
		  ('a subself -> int -> 'a subself) ->
		  ('a subself -> int) *
	          ('a subself -> 'a subself -> real) *
		  ('a subself -> int -> 'a subself)) ->
	         'a subself -> 'a subself
    val pos : 'a subself -> int
    val slide : 'a subself -> int -> 'a subself
    val dist : 'a subself -> 'a subself -> real
end;

functor Place (structure S : EQOBJECT) :> PLACED where Super = S =
struct

structure Super = S;

(* types *)
datatype 'a state = S of
		int *
		(('a subself -> int) *			(* pos   *)
		 ('a subself -> 'a subself -> real) *	(* dist	 *)
		 ('a subself -> int -> 'a subself))	(* slide *)
withtype 'a ext = 'a state * 'a
and 'a subself  = 'a ext Super.subself;
type self = closed subself;

(* messages *)
fun pos p     = let val (_,(S (_,(n,_,_)),_)) = Super.unpack p in n p end;
fun dist p q  = let val (_,(S (_,(_,d,_)),_)) = Super.unpack p in d p q end;
fun slide p n = let val (_,(S (_,(_,_,s)),_)) = Super.unpack p in s p n end;

(* packing *)
fun unpack p = explode (Super.unpack p);
fun repack f p = implode (fn S (s,m) => S (s,f m)) (Super.unpack p);
fun repacks f p = implode (fn S (s,m) => S (f s, m)) (Super.unpack p);
fun mpos p = let val (_,(S (s,_),_)) = Super.unpack p in s end;
fun mslide p n' = repacks (fn n => n + n') p;
fun mdist p q = real (pos q - pos p);
fun peq _ p q = pos p = pos q;
fun pack n = (Super.repack_eq peq, fn ext => (S (n,(mpos,mdist,mslide)),ext));
end;

signature PLACEDOBJ =
sig
    type 'a subself
    type self = closed subself
    val unpack : 'a subself -> ('a -> 'a subself) * 'a
    val repack_eq : (('a subself -> 'a subself -> bool) ->
		     ('a subself -> 'a subself -> bool)) ->
		   'a subself -> 'a subself
    val repack_pos : (('a subself -> int) *
		  ('a subself -> 'a subself -> real) *
		  ('a subself -> int -> 'a subself) ->
		  ('a subself -> int) *
	          ('a subself -> 'a subself -> real) *
		  ('a subself -> int -> 'a subself)) ->
	         'a subself -> 'a subself
end;


(* multidimentional points --------------------------------------------	*)
(* adds an extra coordinate; specializing distance and slide *)

signature XPLACED =
sig
    structure Super : PLACEDOBJ
    type 'a ext
    type 'a subself = 'a ext Super.subself
    type self = closed subself
    val pack : int -> ('a subself -> 'a subself) * ('a -> 'a ext)
    val unpack : 'a subself -> ('a -> 'a subself) * 'a
    val repack : (('a subself -> int) ->
		  ('a subself -> int)) ->
	         'a subself -> 'a subself
    val pos : 'a subself -> int
end;

functor XPlace (structure S : PLACEDOBJ) :> XPLACED where Super = S =
struct

structure Super = S;

(* types *)
datatype 'a state = S of
		int *
		('a subself -> int) 			(* pos	 *)
withtype 'a ext = 'a state * 'a
and 'a subself  = 'a ext Super.subself;
type self = closed subself;

(* messages *)
fun pos p = let val (_,(S (_,n),_)) = Super.unpack p in n p end;

(* packing *)
fun unpack p = explode (Super.unpack p);
fun repack f p = implode (fn S (s,m) => S (s,f m)) (Super.unpack p);
fun repacks f p = implode (fn S (s,m) => S (f s,m)) (Super.unpack p);
fun mpos p = let val (_,(S (s,_),_)) = Super.unpack p in s end;
fun pslide dslide p n = repacks (fn y => y+n) (dslide p n);
fun pdist ddist p q =	sqrt (sq (real (pos q - pos p)) + sq (ddist p q));
fun peq dpeq p q = dpeq p q andalso pos p = pos q;
fun pack n =
	(Super.repack_eq peq o
	 Super.repack_pos (fn (mp,dp,sp) => (mp,pdist dp,pslide sp)),
	 fn ext => (S (n,mpos),ext));
end;



(* colored points  ----------------------------------------------------	*)

signature PAINTED =
sig
    structure Super : EQOBJECT
    type 'a ext
    type 'a subself = 'a ext Super.subself
    type self = closed subself
    val pack : color -> ('a subself -> 'a subself) * ('a -> 'a ext)
    val unpack : 'a subself -> ('a -> 'a subself) * 'a
    val repack : (('a subself -> color) *
		  ('a subself -> color -> 'a subself) ->
		  ('a subself -> color) *
		  ('a subself -> color -> 'a subself)) ->
	         'a subself -> 'a subself
    val col : 'a subself -> color
    val paint : 'a subself -> color -> 'a subself
end;

functor Paint (structure S : EQOBJECT) :> PAINTED where Super = S =
struct

structure Super = S;

(* types *)
datatype 'a state = S of
		color *
		(('a subself -> color) *		(* col	 *)
		 ('a subself -> color -> 'a subself))	(* paint  *)
withtype 'a ext = 'a state * 'a
and 'a subself = 'a ext Super.subself;
type self = closed subself;

(* messages *)
fun col p     = let val (_,(S (_,(c,_)),_)) = Super.unpack p in c p end;
fun paint p c = let val (_,(S (_,(_,pa)),_)) = Super.unpack p in pa p c end;

(* packing *)
fun unpack p = explode (Super.unpack p);
fun repack f p = implode (fn S (s,m) => S (s,f m)) (Super.unpack p);
fun repacks f p = implode (fn S (s,m) => S (f s,m)) (Super.unpack p);
fun mcol p = let val (_,(S (s,_),_)) = Super.unpack p in s end;
fun mpaint p c' = repacks (fn c => c') p;
fun cpeq peq p q = peq p q andalso col p = col q;
fun pack c =
	(Super.repack_eq cpeq, fn ext => (S (c,(mcol,mpaint)),ext));
end;



(* line -------------------------------------------------------	*)

signature MINIPOINT = (* point methods and types *)
    sig
	type 'a subself
	type self = closed subself
	val pos : 'a subself -> int
	val slide : 'a subself -> int -> 'a subself
	val dist : 'a subself -> 'a subself -> real
	val eq : 'a subself -> 'a subself -> bool
    end;

signature LINE =
sig
    structure Param : MINIPOINT
    structure Super : EQOBJECT
    type ('a,'b) ext
    type ('a,'b) subself = ('a,'b) ext Super.subself
    type 'a self = ('a, closed) subself
    val pack : 'a Param.subself * 'a Param.subself ->
	        (('a,'b) subself -> ('a,'b) subself) * ('b -> ('a,'b) ext)
    val unpack : ('a,'b) subself -> ('b -> ('a,'b) subself) * 'b
    val repack : ((('a,'b) subself -> 'a Param.subself) *
		  (('a,'b) subself -> 'a Param.subself) *
		  (('a,'b) subself -> int -> ('a,'b) subself) *
		  (('a,'b) subself -> real) ->
		  (('a,'b) subself -> 'a Param.subself) *
		  (('a,'b) subself -> 'a Param.subself) *
		  (('a,'b) subself -> int -> ('a,'b) subself) *
		  (('a,'b) subself -> real)) ->
		 ('a,'b) subself -> ('a,'b) subself
    val from : ('a,'b) subself -> 'a Param.subself
    val to : ('a,'b) subself -> 'a Param.subself
    val shift : ('a,'b) subself -> int -> ('a,'b) subself
    val length : ('a,'b) subself -> real
end;

functor Join (structure P : MINIPOINT and S : EQOBJECT) :> LINE
	where Super = S and Param = P =
struct

structure Param = P;
structure Super = S;

(* types *)
datatype ('a,'b) state = S of
		('a Param.subself * 'a Param.subself) *
		((('a,'b) subself -> 'a Param.subself) *	(* from *)
		 (('a,'b) subself -> 'a Param.subself) * 	(* to   *)
		 (('a,'b) subself -> int -> ('a,'b) subself) *	(* shift *)
		 (('a,'b) subself -> real))                   	(* length *)
withtype ('a,'b) ext = ('a,'b) state * 'b
and ('a,'b) subself = ('a,'b) ext Super.subself;
type 'a self = ('a,closed) subself;

(* messages *)
fun from l    = let val (_,(S (_,(f,_,_,_)),_)) = Super.unpack l in f l end;
fun to l      = let val (_,(S (_,(_,t,_,_)),_)) = Super.unpack l in t l end;
fun shift l n = let val (_,(S (_,(_,_,s,_)),_)) = Super.unpack l in s l n end;
fun length l  = let val (_,(S (_,(_,_,_,le)),_)) = Super.unpack l in le l end;

(* packing *)
fun unpack p = explode (Super.unpack p);
fun repack f p = implode (fn S (s,m) => S (s,f m)) (Super.unpack p);
fun repacks f p = implode (fn S (s,m) => S (f s,m)) (Super.unpack p);
fun state p = let val (_,(S (s,_),_)) = Super.unpack p in s end;
fun state p = let val (_,(S (s,_),_)) = Super.unpack p in s end;
fun mfr l = let val (f,_) = state l in f end;
fun mto l = let val (_,t) = state l in t end;
fun msh l n = repacks (fn (f,t) => (Param.slide f n, Param.slide t n)) l;
fun mlen l = Param.dist (from l) (to l);
fun leq _ p q =	Param.eq (from p) (from q) andalso Param.eq (to p) (to q);
fun pack (f,t) =
	(Super.repack_eq leq,
	 fn ext => (S ((f,t),(mfr,mto,msh,mlen)), ext));
end;






(* hierarchy ----------------------------------------------------- *)

infix 7 oo;
fun (f,g) oo (F,G) = (f o F, G o g);
fun make (f,g) = f (g %);


structure Eqobj =
    struct
	local
	structure P = Equality (structure S = Root);
	in
	open P;
	fun pack () = P.pack () oo Root.pack ();
	fun new () : self = make (pack ());
	end;
    end;

structure Point =
    struct
	local
	structure P = Place (structure S = struct
					        open Eqobj;
						val repack_eq = Eqobj.repack;
					   end);
	in
	open P;
	fun pack n = P.pack n oo Eqobj.pack ();
	fun new n : self = make (pack n);
	end;
    end;

structure Cpoint =
    struct
	local
	structure P = Paint (structure S = struct
						open Point;
						val repack_eq = Eqobj.repack;
					   end);
	in
	open P;
	fun pack (n,c) = P.pack c oo Point.pack n;
	fun new (n,c) : self = make (pack (n,c));
	end;
    end;

structure D2point =
    struct
	local
	structure P = XPlace (structure S = struct
						open Point;
						val repack_eq = Eqobj.repack;
						val repack_pos =  Point.repack;
					    end);
	in
	open P;
	fun pack (x,y) = P.pack y oo Point.pack x;
	fun new (x,y) : self = make (pack (x,y));
	end;
    end;

structure D3point =
    struct
	local
	structure P = XPlace (structure S = struct
						open D2point;
						val repack_eq = Eqobj.repack;
						val repack_pos =  Point.repack;
					    end);
	in
	open P;
	fun pack (x,y,z) = P.pack z oo D2point.pack (x,y);
	fun new (x,y,z) : self = make (pack (x,y,z));
	end;
    end;

structure Cd3point =
    struct
	local
	structure P = Paint (structure S = struct
						open D3point;
						val repack_eq = Eqobj.repack;
					   end);
	in
	open P;
	fun pack (x,y,z,c) = P.pack c oo D3point.pack (x,y,z);
	fun new (x,y,z,c) : self = make (pack (x,y,z,c));
	end;
    end;

structure Line =
    struct
	local
	structure L = Join (structure P = struct
						open Point;
						val eq = Eqobj.eq;

					   end
			    and S = struct
					open Eqobj;
					val repack_eq = Eqobj.repack;
				    end);
	in
	open L;
	fun pack (f,t) = L.pack (f,t) oo Eqobj.pack ();
	fun new (f,t) : 'a self = make (pack (f,t));
	end;
    end;

